#include "WaveExtractor.hpp"
#include "Logger.hpp"
#include "WaveFileWriter.hpp"
#include <tchar.h>
#include <process.h>

WaveExtractor::WaveExtractor(const Logger* logger)
: file_(NULL), logger_(logger)
{
}

WaveExtractor::~WaveExtractor()
{
	close();
}

bool WaveExtractor::open(const TCHAR* filepath)
{
	PAVIFILE file;
	HRESULT result = AVIFileOpen(&file, filepath, OF_SHARE_DENY_WRITE, NULL);
	if (result == AVIERR_BADFORMAT)
	{
		logger_->error(_T("̓t@CJ܂Bt@CjĂ邩AtH[}bgsłB\n"));
		return false;
	}
	else if (result == AVIERR_MEMORY)
	{
		logger_->error(_T("̓t@CJ܂BȂȂ܂B\n"));
		return false;
	}
	else if (result == AVIERR_FILEREAD)
	{
		logger_->error(_T("̓t@CJ܂BǂݍݒɃfBXNG[܂B\n"));
		return false;
	}
	else if (result == AVIERR_FILEOPEN)
	{
		logger_->error(_T("̓t@CJ܂Bt@CJۂɃfBXNG[܂B\n"));
		return false;
	}
	else if (result == REGDB_E_CLASSNOTREG)
	{
		logger_->error(_T("̓t@CJ܂Bw肳ꂽt@C^CvnhAWXgɌ܂łB\n"));
		return false;
	}
	else if (FAILED(result))
	{
		logger_->error(_T("̓t@CJ܂BsȃG[܂B\n"));
		return false;
	}

	file_ = file;

	logger_->info(_T("̓t@CJ܂B\n"));

	return true;
}

void WaveExtractor::close()
{
	if (file_ != NULL)
	{
		AVIFileRelease(file_);
		file_ = NULL;
		logger_->info(_T("̓t@C܂B\n"));
	}
}

bool WaveExtractor::extract(int streamNumber, const TCHAR* outputFilePath)
{
	// w̃I[fBIXg[J
	PAVISTREAM stream;
	HRESULT result = AVIFileGetStream(file_, &stream, streamtypeAUDIO, streamNumber);
	if (result == AVIERR_NODATA)
	{
		logger_->error(_T("I[fBIXg[J܂Bw肳ꂽXg[݂͑܂B\n"));
		return false;
	}
	else if (result == AVIERR_MEMORY)
	{
		logger_->error(_T("I[fBIXg[J܂BȂȂ܂B\n"));
		return false;
	}
	else if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[J܂BsȃG[܂B\n"));
		return false;
	}

	logger_->info(_T("I[fBIXg[J܂B\n"));

	// o{
	bool success = extractImpl(stream, outputFilePath);

	// I[fBIXg[
	AVIStreamRelease(stream);

	logger_->info(_T("I[fBIXg[܂B\n"));

	return success;
}

bool WaveExtractor::extractImpl(PAVISTREAM stream, const TCHAR* outputFilePath)
{
	// I[fBIXg[̏擾
	AVISTREAMINFO  streamInfo;
	HRESULT result = AVIStreamInfo(stream, &streamInfo, sizeof(AVISTREAMINFO));
	if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[̎擾Ɏs܂B\n"));
		return false;
	}

	logger_->info(_T("I[fBIXg[擾܂B\n"));
	logger_->verbos(_T("TṽTCY: %d Byte\n"), streamInfo.dwSampleSize);
	logger_->verbos(_T("Tv: %d\n"), streamInfo.dwLength);


	// I[fBIXg[WavetH[}bg擾
	LONG formatSize = 0;
	result = AVIStreamReadFormat(stream, 0, NULL, &formatSize);
	if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[WavetH[}bg擾Ɏs܂B\n"));
		return false;
	}
	if (formatSize > sizeof(WAVEFORMATEX))
	{
		logger_->error(_T("I[fBIXg[WavetH[}bg擾Ɏs܂B\n"));
		return false;
	}

	WAVEFORMATEX format;
	ZeroMemory(&format, sizeof(WAVEFORMATEX));
	result = AVIStreamReadFormat(stream, 0, &format, &formatSize);
	if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[WavetH[}bg擾Ɏs܂B\n"));
		return false;
	}

	logger_->info(_T("I[fBIXg[WavetH[}bg擾܂B\n"));
	logger_->verbos(_T("`l: %d\n"), format.nChannels);
	logger_->verbos(_T("rbg: %d\n"), format.wBitsPerSample);
	logger_->verbos(_T("TvO[g: %d\n"), format.nSamplesPerSec);


	// g`f[^ǂݍWavet@CƂĕۑ
	bool complete = false; // tO

	size_t bufferNumber = 0;
	size_t bufferLength = streamInfo.dwSampleSize * 2 * 1024 * 1024;
	BYTE* buffer = new BYTE[bufferLength * 2];

	WaveFileWriter writer(format, streamInfo.dwSampleSize * streamInfo.dwLength, logger_);
	if (writer.open(outputFilePath))
	{
		DWORD cursor = 0;
		for (;;)
		{
			// ǂݍ݊𔻒
			if (cursor >= streamInfo.dwLength)
			{
				complete = true;
				break;
			}

			// gpobt@ԍ
			bufferNumber ^= 1;

			// ǂݏoTv擾
			size_t readSamples = bufferLength / streamInfo.dwSampleSize;
			if (readSamples > streamInfo.dwLength - cursor)
				readSamples = streamInfo.dwLength - cursor;

			// ǂݍݏ
			LONG writeBytes;
			LONG writeSamples;
			result = AVIStreamRead(stream, cursor, static_cast<LONG>(readSamples), &buffer[bufferLength * bufferNumber], static_cast<LONG>(bufferLength), &writeBytes, &writeSamples);
			if (result == AVIERR_BUFFERTOOSMALL)
			{
				logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂Bǂݍ݃obt@̃TCY܂B\n"));
				break;
			}
			else if (result == AVIERR_MEMORY)
			{
				logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂BȂȂ܂B\n"));
				break;
			}
			else if (result == AVIERR_FILEREAD)
			{
				logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂BfBXNG[܂B\n"));
				break;
			}
			else if (FAILED(result))
			{
				logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂BsȃG[܂B\n"));
				break;
			}

			logger_->debug(_T("%d - %d ̃f[^ǂݎ܂B\n"), cursor * streamInfo.dwSampleSize, (cursor + writeSamples) * streamInfo.dwSampleSize - 1);

			// vƈႤʂǂݍ܂ꂽG[
			// G[ƂfBXNG[͏ŃJo[Ă̂ōȂR͖͂
			if (readSamples != writeSamples || streamInfo.dwSampleSize * writeSamples != writeBytes)
			{
				logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂Bf[^TCYɕs܂B\n"));
				break;
			}

			// ǂݍ݃J[\XV
			cursor += writeSamples;

			// o҂
			if (!writer.returnWhenWritten())
				break;

			// ̏o
			if (!writer.write(&buffer[bufferLength * bufferNumber], writeBytes))
				break;
		}

		writer.close();
	}

	delete [] buffer;

	return complete;
}
